/**
 *  Collect simple execution statistics.
 *
 *  @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 *
 */

#ifndef WLPDSTM_STATS_H_
#define WLPDSTM_STATS_H_

#include <cstring>
#include <stdint.h>
#include <iostream>

#include "common/cache_aligned_alloc.h"
#include "constants.h"

namespace wlpdstm {

	class Statistics : public CacheAlignedAlloc {
	public:
		enum Type {
			COMMIT = 0,
			COMMIT_READ_ONLY,
			ABORT,
			ABORT_WRITE_LOCKED,
			ABORT_WRITE_VALIDATE,
			ABORT_READ_VALIDATE,
			ABORT_READ_LOCKED,
			ABORT_COMMIT_VALIDATE,
			CLOCK_OVERFLOWS,
#ifdef DETAILED_STATS
			EXTEND_SUCCESS,
			EXTEND_FAILURE,
			FREE_DOMINATED,
			UPDATE_MINIMUM_TS,
			SWITCH_TO_SECOND_CM_PHASE,
			CM_DECIDE,
			LARGER_TABLE_WRITE_HITS,
			LARGER_TABLE_READ_HITS,
			WRITES,
			NEW_WRITES,
			READS,
			NEW_READS,
			FALSE_CONFLICTS,
			LOCK_TABLE_RECONFIGURATIONS,
//			READ_AFTER_WRITE,
//			WRITE_AFTER_WRITE,
//			VALIDATE_READ_SET_SIZE,
			MEMORY_DEALLOC_COUNT,
			MEMORY_DEALLOC_SIZE,
			MEMORY_DEALLOC_LOCKS,
			WAIT_ON_ABORT,
#endif /* DETAILED_STATS */
			COUNT
		};

		static const char *GetTypeName(Type type) {
			const char *type_names[] = {
				"Commit",
				"CommitReadOnly",
				"Abort",
				"AbortWriteLocked",
				"AbortWriteValidate",
				"AbortReadValidate",
				"AbortReadLocked",
				"AbortCommitValidate",
				"ClockOverflows",
#ifdef DETAILED_STATS
				"ExtendSuccess",
				"ExtendFailure",
				"FreeDominated",
				"UpdateMinimumTs",
				"SwitchToSecondCMPhase",
				"CMDecide",
				"LargerTableWriteHits",
				"LargerTableReadHits",
				"Writes",
				"NewWrites",
				"Reads",
				"NewReads",
				"FalseConflicts",
				"LockTableReconfigurations",
//				"ReadAfterWrite",
//				"WriteAfterWrite",
//				"ValidateReadSetSize",
				"MemoryDeallocCount",
				"MemoryDeallocSize",
				"MemoryDeallocLocks",
				"WaitOnAbort"
#endif /* DETAILED_STATS */
			};

			return type_names[type];
		}
		
	public:
		Statistics() {
			Reset();
		}

		void Reset() {
			memset(&stat_counts, 0, sizeof(stat_counts));
		}

		void Increment(Type stat, int inc = 1) {
			stat_counts[stat] += inc;
		}

		void Merge(Statistics *other) {
			for(unsigned i = 0;i < COUNT;i++) {
				Increment((Type)i, other->Get((Type)i));
			}			
		}

		uint32_t Get(Type stat) {
			return stat_counts[stat];
		}

	private:
		uint32_t stat_counts[COUNT];
	};

	// hold statistics for a single thread and for lexical transactions
	struct ThreadStatistics {
		static const int MAX_LEXICAL_STATS = 64;

		ThreadStatistics() : stats(new Statistics()), next(NULL), lexical_tx_id(NO_LEXICAL_TX) {
			for(int i = 0;i < MAX_LEXICAL_STATS;i++) {
				lexical_tx_stats[i] = new Statistics();
				lexical_tx_stat_used[i] = false;
			}
		}

		void Merge(const ThreadStatistics *tstat) {
			stats->Merge(tstat->stats);

			for(int i = 0;i < MAX_LEXICAL_STATS;i++) {
				lexical_tx_stats[i]->Merge(tstat->lexical_tx_stats[i]);
				lexical_tx_stat_used[i] |= tstat->lexical_tx_stat_used[i];
			}
		}

		inline void IncrementStatistics(Statistics::Type type, unsigned inc = 1) {
#ifdef COLLECT_STATS
			stats->Increment(type, inc);

			if(lexical_tx_id != -1) {
				lexical_tx_stats[lexical_tx_id]->Increment(type, inc);
				lexical_tx_stat_used[lexical_tx_id] = true;
			}
#endif
		}
		
		void Print(std::ostream &out_stream, unsigned indent = 1);
		void PrintIndent(std::ostream &out_stream, unsigned indent);

		Statistics *stats;

		Statistics *lexical_tx_stats[MAX_LEXICAL_STATS];
		bool lexical_tx_stat_used[MAX_LEXICAL_STATS];

		// this is used just for finalization
		ThreadStatistics *next;

		int lexical_tx_id;
	};

	// class for manipulating group of statistics
	class ThreadStatisticsCollection {
		public:
			ThreadStatisticsCollection() : head(NULL), last(NULL) {
				// empty
			}

			void Add(ThreadStatistics *stat) {
				if(last == NULL) {
					head = last = stat;
				} else {
					last->next = stat;
					last = stat;
				}
			}

			ThreadStatistics MergeAll() {
				ThreadStatistics ret;
				ThreadStatistics *curr;

				for(curr = head;curr != NULL;curr = curr->next) {
					ret.Merge(curr);
				}

				return ret;
			}

		private:
			ThreadStatistics *head;
			ThreadStatistics *last;
	};
}

inline void wlpdstm::ThreadStatistics::PrintIndent(std::ostream &out_stream, unsigned indent) {	
	for(unsigned u = 0;u < indent;u++) {
		out_stream << "\t";
	}
}
		
inline void wlpdstm::ThreadStatistics::Print(std::ostream &out_stream, unsigned indent) {	
	for(unsigned i = 0;i < Statistics::COUNT;i++) {
		PrintIndent(out_stream, indent);
		out_stream << Statistics::GetTypeName((Statistics::Type)i) << ": "
			<< stats->Get((Statistics::Type)i) << std::endl;
	}

	for(int ls = 0;ls < MAX_LEXICAL_STATS;ls++) {
		if(lexical_tx_stat_used[ls]) {
			PrintIndent(out_stream, indent);
			out_stream << "Tx " << ls << ":" << std::endl;

			for(unsigned i = 0;i < Statistics::COUNT;i++) {
				PrintIndent(out_stream, indent + 1);
				out_stream << Statistics::GetTypeName((Statistics::Type)i) << ": "
				<< lexical_tx_stats[ls]->Get((Statistics::Type)i) << std::endl;
			}
		}
	}
}

#endif /* WLPDSTM_STATS_H_ */
