/**
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#include <cstdlib>
#include <stdint.h>
#include <cstring>

#include "aborted_exc.h"
#include "common/version_lock.h"
#include "constants.h"
#include "common/log.h"
#include "common/timestamp.h"
#include "memory.h"
#include "tid.h"
#include "common/tls.h"
#include "common/cache_aligned_alloc.h"
#include "common/jmp.h"
#include "stats.h"
#include "common/large_lock_set.h"
#include "common/timing.h"
#include "common/random.h"

#ifdef USE_PREEMPTIVE_WAITING
#include "common/preemptive_utils.h"
#endif // USE_PREEMPTIVE_WAITING

// TODO: perform sensitivity analysis here as well
// default is 6 0

// how many successive locations (semgent size) does one lock cover (log)
//#define LOCK_EXTENT_SIZE 6
#define DEFAULT_LOCK_EXTENT_SIZE 4
#define MIN_LOCK_EXTENT_SIZE 2
#define MAX_LOCK_EXTENT_SIZE 10
// how many segments are covered with a single lock
//#define LOCK_OVERLOADING 0 

//#define VERSION_LOCK_TABLE_SIZE (1 << (ADDRESS_SPACE_SIZE - LOCK_EXTENT_SIZE - LOCK_OVERLOADING))

#define VERSION_LOCK_TABLE_SIZE (1 << 22)
// two locks are used - write/write and read/write 
#define FULL_VERSION_LOCK_TABLE_SIZE (VERSION_LOCK_TABLE_SIZE << 1)

#ifdef ADAPTIVE_LOCKING
#define CHANGE_TABLE_CONFIGURATION_FREQUENCY 100
#endif /* ADAPTIVE_LOCKING */

// TODO: perform sensitivity analysis for this parameter
// rbtree
//     20 no tx go to second phase
//     10 several tx fo to second phase
//      5 significant number of tx go to second phase
// default is 10
#define CM_ACCESS_THRESHOLD 10

// TODO: perform sensitivity analysis for this parameter
// rbtree
//     20 number of FreeDominated calls is small and this parameter does not make difference
// default is 20
#define UPDATE_LOCAL_LAST_OBSERVED_TS_FREQUENCY 20

namespace wlpdstm {
	
	union PaddedUnsigned {
		volatile unsigned val;
		char padding[CACHE_LINE_SIZE_BYTES];
	};

	union PaddedBool {
		volatile bool val;
		char padding[CACHE_LINE_SIZE_BYTES];
	};	

	typedef VersionLock WriteLock;

	class TxMixinv : public CacheAlignedAlloc {

		//////////////////////////////////
		// log related structures start //
		//////////////////////////////////

		struct ReadLogEntry {
			VersionLock *read_lock;
			VersionLock version;
		};

		struct WriteWordLogEntry {
			void *address;
			void *value;
			WriteWordLogEntry *next;
		};

#ifdef SUPPORT_LOCAL_WRITES
		struct WriteWordLocalLogEntry {
			void *address;
			void *value;
		};
#endif /* SUPPORT_LOCAL_WRITES */

		struct WriteLogEntry {
			VersionLock *read_lock;
			WriteLock *write_lock;

			VersionLock old_version;

			TxMixinv *owner;

#ifdef WORD_LOG_HASHTABLE
			static const unsigned HASHTABLE_SIZE = 16;
			static const unsigned HASH_MASK = (HASHTABLE_SIZE - 1) << WORD_LOG_SIZE;
			WriteWordLogEntry table[HASHTABLE_SIZE];
#else
			WriteWordLogEntry *head;
#endif /* WORD_LOG_HASHTABLE */

			// methods
			void InsertWordLogEntry(void *address, void *value);

			WriteWordLogEntry *FindWordLogEntry(void *address);

			void ClearWordLogEntries();

#ifdef WORD_LOG_HASHTABLE
			unsigned HashAddress(uintptr_t address) {
				return (address & HASH_MASK) >> WORD_LOG_SIZE;
			}
#endif /* WORD_LOG_HASHTABLE */
		};

		typedef Log<ReadLogEntry> ReadLog;
		typedef Log<WriteLogEntry> WriteLog;
#ifdef SUPPORT_LOCAL_WRITES
		typedef Log<WriteWordLocalLogEntry> WriteLocalLog;
#endif /* SUPPORT_LOCAL_WRITES */
		typedef Log<WriteWordLogEntry> WriteWordLogMemPool;

		////////////////////////////////
		// log related structures end //
		////////////////////////////////

	public:
		// possible CM phases
		enum CmPhase {
			CM_PHASE_INITIAL,
			CM_PHASE_GREEDY
		};

		enum RestartCause {
			RESTART_EXTERNAL,
			RESTART_LOCK,
			RESTART_VALIDATION
		};

	public:
		static void GlobalInit();

		static void InitializeReadLocks();

		static void InitializeWriteLocks();
		
		void ThreadInit();
		
		void TxStart(int lex_tx_id = NO_LEXICAL_TX);
		
		void TxCommit();
		
		void TxAbort();
		
		void TxAbortFull();
		
		void WriteWord(void *address, void *val);

#ifdef SUPPORT_LOCAL_WRITES
		void WriteWordLocal(void *address, void *val);
#endif /* SUPPORT_LOCAL_WRITES */
		
		void *ReadWord(void *address);
		
		void *TxMalloc(size_t size);
		
		void TxFree(void *ptr, size_t size);
		
		void TxRestart(RestartCause cause = RESTART_EXTERNAL);

	private:
		void *ReadWordInner(void *address);

		void RestartJump();

		void AbortJump();
		
		void LockMemoryBlock(void *address, size_t size);

		unsigned map_address_to_index(void *address, unsigned le);

		unsigned map_address_to_index(void *address);
		
		VersionLock *map_address_to_read_lock(void *address);
		
		WriteLock *map_address_to_write_lock(void *address);
		
		WriteLock *map_read_lock_to_write_lock(VersionLock *lock_address);

		VersionLock *map_write_lock_to_read_lock(WriteLock *lock_address);
		
		WriteLogEntry *LockMemoryStripe(WriteLock *write_lock, void *address);
		
		void Rollback();
		
		bool Validate();

		bool ValidateCommit();

		void ReleaseReadLocks();
		
		bool Extend();

		bool LockedByMe(WriteLogEntry *log_entry);
		
		///////////////////////////
		// contention management //
		///////////////////////////
		
		bool ShouldAbortWrite(WriteLock *write_lock);
		
		// TODO: isolate CM in a separate class
		bool CMStrongerThan(TxMixinv *other);
		
		void CmStartTx();
		
		void CmOnAccess();

#ifdef WAIT_ON_SUCC_ABORTS
		void WaitOnAbort();
#endif /* WAIT_ON_SUCC_ABORTS */

		// last observed ts
		void InitLastObservedTs();
		
		void UpdateLastObservedTs(unsigned ts);
		
		void Register();

		void YieldCPU();
		
	public:
		static unsigned UpdateMinimumObservedTs();
		
		static unsigned GetMinimumObservedTs();
		
	private:
		// local
		ReadLog read_log;
		
		// local
		WriteLog write_log;

#ifdef SUPPORT_LOCAL_WRITES
		// local
		WriteLocalLog write_local_log;
#endif /* SUPPORT_LOCAL_WRITES */

		// local
		WriteWordLogMemPool write_word_log_mem_pool;
		
		// local
		unsigned valid_ts;

		// contention management
		// local
		unsigned locations_accessed;
		
		// local
		CmPhase cm_phase;

#ifdef WAIT_ON_SUCC_ABORTS
		const static unsigned SUCC_ABORTS_THRESHOLD = 1;
		const static unsigned SUCC_ABORTS_MAX = 10;

		const static unsigned WAIT_CYCLES_MULTIPLICATOR = 8000;
#endif /* WAIT_ON_SUCC_ABORTS */

		unsigned succ_aborts;

		// memory management support
		// local
		unsigned start_count;
		
		//////////////////////
		// shared variables //
		//////////////////////
		
		static VersionLock version_lock_table[FULL_VERSION_LOCK_TABLE_SIZE];
		
		static GlobalTimestamp commit_ts;
		
		static GlobalTimestamp cm_ts;
		
		static PaddedUnsigned minimum_observed_ts;
		
		static PaddedSpinTryLock minimum_observed_ts_lock;
		
		static TxMixinv *transactions[MAX_THREADS];
		
		static unsigned thread_count;
		
	public:
		// local
		Tid tid;
		
		// local
		MemoryManager mm;
		
	private:
		// data accessed by other tx
		// ro shared
		CACHE_LINE_ALIGNED union {
			volatile unsigned greedy_ts;
			char padding[CACHE_LINE_SIZE_BYTES];
		} aligned_greedy_ts;
		
		// w shared
		CACHE_LINE_ALIGNED union {
			volatile bool aborted_externally;
			char padding[CACHE_LINE_SIZE_BYTES];
		} aligned_aborted_externally;
		
		// r shared
		CACHE_LINE_ALIGNED union {
			volatile unsigned last_observed_ts;
			char padding[CACHE_LINE_SIZE_BYTES];
		} aligned_last_observed_ts;
		
		// stats
	private:
		// local
		ThreadStatistics stats;

	public:
		static void PrintStatistics();

#ifdef USE_LONG_JUMP
		// local
		LONG_JMP_BUF start_buf;
		
#endif /* USE_LONG_JUMP */

		////////////////////////////////////////
		// synchronize all transactions start //
		////////////////////////////////////////

	private:
		volatile bool executing;

		static PaddedBool synchronization_in_progress;

		bool StartSynchronization();

		void EndSynchronization();

		bool Synchronize();

		void RestartCommitTS();

		//////////////////////////////////////
		// synchronize all transactions end //
		//////////////////////////////////////

		/////////////////////////////////
		// lock mapping function start //
		/////////////////////////////////

#ifdef ADAPTIVE_LOCKING
		// currently used lock extent
		static PaddedUnsigned lock_extent;

		// main parameters of the algorithm
		const static unsigned LOCK_EXTENT_INCREASE_THRESHOLD = 2;
		const static unsigned LOCK_EXTENT_DECREASE_THRESHOLD = 60;

		// local
		LargeLockSet<VERSION_LOCK_TABLE_SIZE> larger_lock_extent_table;

		// local
		// this is lock_extend for which copy of the table is maintained
		unsigned larger_lock_extent;

		const static unsigned NOT_PROPOSING_LOCK_EXTENT = MIN_LOCK_EXTENT_SIZE - 1;
		const static unsigned LOCK_EXTENT_CHANGED = MAX_LOCK_EXTENT_SIZE + 1;
		const static unsigned SIGNIFICANT_STAT_COUNT_ABORTS = 100;
		const static unsigned SIGNIFICANT_STAT_COUNT_WRITES = 100;

		// shared
		// proposing new lock extent
		CACHE_LINE_ALIGNED PaddedUnsigned proposed_lock_extent;

		unsigned old_lock_extent;

		// local - statistics
		unsigned writes;
		unsigned new_writes;
		unsigned larger_table_hits;
		unsigned larger_table_read_hits;
		unsigned aborts;
		unsigned false_sharing;
		const static unsigned FALSE_SHARING_HISTORY_SIZE = MAX_LOCK_EXTENT_SIZE - MIN_LOCK_EXTENT_SIZE + 1;
		static unsigned false_sharing_history[FALSE_SHARING_HISTORY_SIZE];

		unsigned reads;

		// local
		unsigned change_table_configuration_counter;

		// helper functions
		bool IsFalseConflict(WriteLogEntry *log_entry, void *address);
		void ClearLocalStatistics();

		// if table configuration changed apply it
		void ApplyTableConfigurationChange();

		// propose table change if necessary
		void ProposeTableChange();
		void ProposeTableChangeInner();

		bool EnoughResizeTableDataAborts();
		bool EnoughResizeTableDataWrites();

		unsigned CalculateFalseAbortRate();
		unsigned CalculateLargerTableHitRate();
//		static unsigned GetFalseSharingHistoryIndex(unsigned le);
		static unsigned GetFalseSharingHistory(unsigned le);

#endif /* ADAPTIVE_LOCKING */

		///////////////////////////////
		// lock mapping function end //
		///////////////////////////////

		///////////////////////////
		// measure timings start //
		///////////////////////////

#ifdef MEASURE_TIMING
		static const unsigned MEASURE_TIMING_FREQUENCY = 100;

		unsigned measure_write_timing_count;
		unsigned measure_read_timing_count;

		unsigned write_timing_count;
		uint64_t write_timing_total;

		unsigned read_timing_count;
		uint64_t read_timing_total;

		bool ShouldMeasureWriteTiming();
		bool ShouldMeasureReadTiming();

		void ResetTimingData();

		void PrintTimingData();
#endif /* MEASURE_TIMING */

		/////////////////////////
		// measure timings end //
		/////////////////////////

		Random random;

		// padding of one cache line
		volatile char padding[CACHE_LINE_SIZE_BYTES];

	} __attribute__ ((aligned(CACHE_LINE_SIZE_BYTES)));
}

#ifdef ADAPTIVE_LOCKING
#define LOCK_EXTENT lock_extent.val
#else
#define LOCK_EXTENT DEFAULT_LOCK_EXTENT_SIZE
#endif /* ADAPTIVE_LOCKING */

//////////////////////////
// initialization start //
//////////////////////////

inline void wlpdstm::TxMixinv::GlobalInit() {
	InitializeReadLocks();

	InitializeWriteLocks();

	// initialize shared data
	minimum_observed_ts.val = MINIMUM_TS;
	
	// initialize memory manager
	MemoryManager::GlobalInit();

	synchronization_in_progress.val = false;

#ifdef ADAPTIVE_LOCKING
	lock_extent.val = DEFAULT_LOCK_EXTENT_SIZE;

	for(unsigned i = 0;i < FALSE_SHARING_HISTORY_SIZE;i++) {
		false_sharing_history[i] = 0;
	}
#endif /* ADAPTIVE_LOCKING */
}

inline void wlpdstm::TxMixinv::InitializeReadLocks() {
	VersionLock initial_version = get_version_lock(MINIMUM_TS + 1);
	
	for(int i = 0;i < FULL_VERSION_LOCK_TABLE_SIZE;i += 2) {
		version_lock_table[i] = initial_version;
	}	
}

inline void wlpdstm::TxMixinv::InitializeWriteLocks() {
	for(int i = 1;i < FULL_VERSION_LOCK_TABLE_SIZE;i += 2) {
		version_lock_table[i] = WRITE_LOCK_CLEAR;
	}	
}

inline void wlpdstm::TxMixinv::ThreadInit() {
	aligned_aborted_externally.aborted_externally = false;
	
	InitLastObservedTs();
	
	Register();
	
	// locally initialize memory manager
	mm.ThreadInit(tid.Get());

#ifdef COLLECT_STATS
	mm.InitStats(&stats);
#endif

	executing = false;

#ifdef ADAPTIVE_LOCKING
	// init change lock table conf
	proposed_lock_extent.val = NOT_PROPOSING_LOCK_EXTENT;
	ClearLocalStatistics();
	change_table_configuration_counter = 0;
	larger_lock_extent = lock_extent.val + 1;
	old_lock_extent = lock_extent.val;
#endif /* ADAPTIVE_LOCKING */

#ifdef MEASURE_TIMING
	ResetTimingData();
#endif /* MEASURE_TIMING */

	succ_aborts = 0;
}

////////////////////////
// initialization end //
////////////////////////


////////////////////////////
// mapping to locks start //
////////////////////////////

inline unsigned wlpdstm::TxMixinv::map_address_to_index(void *address) {
	return map_address_to_index(address, LOCK_EXTENT) << 1;
}

inline unsigned wlpdstm::TxMixinv::map_address_to_index(void *address, unsigned le) {
	return (((uintptr_t)address >> (le)) & (VERSION_LOCK_TABLE_SIZE - 1));
}

inline wlpdstm::VersionLock *wlpdstm::TxMixinv::map_address_to_read_lock(void *address) {
	return version_lock_table + map_address_to_index(address);
}

inline wlpdstm::WriteLock *wlpdstm::TxMixinv::map_address_to_write_lock(void *address) {
	return map_address_to_read_lock(address) + 1;
}

inline wlpdstm::WriteLock *wlpdstm::TxMixinv::map_read_lock_to_write_lock(VersionLock *lock_address) {
	return lock_address + 1;
}

inline wlpdstm::VersionLock *wlpdstm::TxMixinv::map_write_lock_to_read_lock(WriteLock *lock_address) {
	return lock_address - 1;
}

//////////////////////////
// mapping to locks end //
//////////////////////////


//////////////////////////
// main algorithm start //
//////////////////////////

inline void wlpdstm::TxMixinv::TxStart(int lex_tx_id) {
	if(Synchronize()) {
#ifdef ADAPTIVE_LOCKING
		ApplyTableConfigurationChange();
#endif /* ADAPTIVE_LOCKING */
	}

	mm.TxStart();

	// get validity timestamp
	valid_ts = commit_ts.readCurrentTs();

	UpdateLastObservedTs(valid_ts);
	
	CmStartTx();
	
	// reset aborted flag
	aligned_aborted_externally.aborted_externally = false;

	// initialize lexical tx id
	stats.lexical_tx_id = lex_tx_id;

	atomic_store_release(&executing, true);

#ifdef ADAPTIVE_LOCKING
	// reset table
	larger_lock_extent_table.Clear();
#endif /* ADAPTIVE_LOCKING */
}

inline void wlpdstm::TxMixinv::TxCommit() {
	unsigned ts = MINIMUM_TS;
	bool read_only = write_log.empty();

	if(!read_only) {
		// first lock all read locks
		for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
			WriteLogEntry &entry = *iter;
			*(entry.read_lock) = READ_LOCK_SET;
		}

		// now get a commit timestamp with full barrier
		ts = commit_ts.getNextTsFull() + 1;

		// if global time overflows restart
		if(ts >= MAXIMUM_TS) {
			atomic_store_no_barrier(&executing, false);
			ReleaseReadLocks();
			Rollback();

			if(StartSynchronization()) {
				RestartCommitTS();
				EndSynchronization();
				stats.IncrementStatistics(Statistics::CLOCK_OVERFLOWS);
			}

			RestartJump();
		}

		if(ts != valid_ts + 1 && !ValidateCommit()) {
			ReleaseReadLocks();
			stats.IncrementStatistics(Statistics::ABORT_COMMIT_VALIDATE);
#ifdef ADAPTIVE_LOCKING
			++aborts;
#endif /* ADAPTIVE_LOCKING */
			TxRestart(RESTART_VALIDATION);
		}

		VersionLock commitVersion = get_version_lock(ts);

		// now update all written values
		for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
			WriteLogEntry &entry = *iter;

			// now update actual values
#ifdef WORD_LOG_HASHTABLE
			WriteWordLogEntry *word_log_entry;

			for(unsigned i = 0;i < WriteLogEntry::HASHTABLE_SIZE;i++) {
				if(entry.table[i].address != NULL) {
					*(void **)entry.table[i].address = entry.table[i].value;
					word_log_entry = entry.table[i].next;
#else
			WriteWordLogEntry *word_log_entry = entry.head;
#endif /* WORD_LOG_HASHTABLE */
					while(word_log_entry != NULL) {
						*(void **)word_log_entry->address =	word_log_entry->value;
						word_log_entry = word_log_entry->next;
					}

#ifdef WORD_LOG_HASHTABLE
				}
			}
#endif /* WORD_LOG_HASHTABLE */
			
			// release locks
			atomic_store_release(entry.read_lock, commitVersion);
			atomic_store_release(entry.write_lock, WRITE_LOCK_CLEAR);
		}

		write_log.clear();
		write_word_log_mem_pool.clear();
	} else {
#ifdef DETAILED_STATS
		stats.IncrementStatistics(Statistics::COMMIT_READ_ONLY);
#endif /* DETAILED_STATS */
	}

	read_log.clear();
#ifdef SUPPORT_LOCAL_WRITES
	write_local_log.clear();
#endif /* SUPPORT_LOCAL_WRITES */
	
	// commit mem
	mm.TxCommit<TxMixinv>(ts);
	
	stats.IncrementStatistics(Statistics::COMMIT);

	atomic_store_release(&executing, false);

	succ_aborts = 0;
	
#ifdef ADAPTIVE_LOCKING
	if(!read_only) {
		ProposeTableChange();
	}
#endif /* ADAPTIVE_LOCKING */
}

inline void wlpdstm::TxMixinv::ReleaseReadLocks() {
	for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
		WriteLogEntry &entry = *iter;
		*(entry.read_lock) = entry.old_version;
	}			
}

// updates only write locks
inline void wlpdstm::TxMixinv::Rollback() {
#ifdef SUPPORT_LOCAL_WRITES
	// rollback local writes
	for(WriteLocalLog::iterator iter = write_local_log.begin();iter.hasNext();iter.next()) {
		WriteWordLocalLogEntry &entry = *iter;
		*(void **)entry.address = entry.value;
	}
#endif /* SUPPORT_LOCAL_WRITES */

	// drop locks
	for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
		WriteLogEntry &entry = *iter;
		*entry.write_lock = WRITE_LOCK_CLEAR;
	}
	
	// empty logs
	read_log.clear();
	write_log.clear();
#ifdef SUPPORT_LOCAL_WRITES
	write_local_log.clear();
#endif /* SUPPORT_LOCAL_WRITES */
	write_word_log_mem_pool.clear();

	YieldCPU();
	
	mm.TxAbort();

	stats.IncrementStatistics(Statistics::ABORT);
}

inline wlpdstm::TxMixinv::WriteLogEntry *wlpdstm::TxMixinv::LockMemoryStripe(WriteLock *write_lock, void *address) {
#ifdef DETAILED_STATS
	stats.IncrementStatistics(Statistics::WRITES);
#endif /* DETAILED_STATS */
#ifdef ADAPTIVE_LOCKING
	++writes;
#endif /* ADAPTIVE_LOCKING */

	// read lock value
	WriteLock lock_value = (WriteLock)atomic_load_no_barrier(write_lock);
	bool locked = is_write_locked(lock_value);

	if(locked) {
		WriteLogEntry *log_entry = (WriteLogEntry *)lock_value;
		
		if(LockedByMe(log_entry)) {
			return log_entry;
		}
	}

#ifdef DETAILED_STATS
	stats.IncrementStatistics(Statistics::NEW_WRITES);
#endif /* DETAILED_STATS */
#ifdef ADAPTIVE_LOCKING
	++new_writes;
	bool false_conflict_detected = false;
#endif /* ADAPTIVE_LOCKING */

	while(true) {		
		if(locked) {
#ifdef ADAPTIVE_LOCKING
			if(!false_conflict_detected) {
				false_conflict_detected = true;
				
				if(IsFalseConflict((WriteLogEntry *)lock_value, address)) {
					stats.IncrementStatistics(Statistics::FALSE_CONFLICTS);
					++false_sharing;
				}
			}
#endif /* ADAPTIVE_LOCKING */
			
			if(ShouldAbortWrite(write_lock)) {
				stats.IncrementStatistics(Statistics::ABORT_WRITE_LOCKED);
#ifdef ADAPTIVE_LOCKING
				++aborts;
#endif /* ADAPTIVE_LOCKING */

//				if(IsFalseConflict((WriteLogEntry *)lock_value, address)) {
//					stats.IncrementStatistics(Statistics::FALSE_CONFLICTS);
//					++false_sharing;
//				}
				
				TxRestart(RESTART_LOCK);
			} else {
				lock_value = (WriteLock)atomic_load_acquire(write_lock);
				locked = is_write_locked(lock_value);
				YieldCPU();
			}
		}

		// prepare write log entry
		WriteLogEntry *log_entry = write_log.get_next();
		log_entry->write_lock = write_lock;
		log_entry->ClearWordLogEntries(); // need this here TODO - maybe move this to commit/abort time
		log_entry->owner = this; // this is for CM - TODO: try to move it out of write path

		// now try to lock it
		if(atomic_cas_release(write_lock, WRITE_LOCK_CLEAR, log_entry)) {
			// need to check read set validity if this address was read before
			// we skip this read_before() check TODO: maybe do that
			VersionLock *read_lock = map_write_lock_to_read_lock(write_lock);
			VersionLock version = (VersionLock)atomic_load_acquire(read_lock);

			if(get_value(version) > valid_ts) {
				if(!Extend()) {
					stats.IncrementStatistics(Statistics::ABORT_WRITE_VALIDATE);
#ifdef ADAPTIVE_LOCKING
					++aborts;
#endif /* ADAPTIVE_LOCKING */
					TxRestart(RESTART_VALIDATION);
				}
			}

			// success
			log_entry->read_lock = read_lock;
			log_entry->old_version = version;
			CmOnAccess();

#ifdef ADAPTIVE_LOCKING
			if(larger_lock_extent_table.Set(map_address_to_index(address, larger_lock_extent))) {
				stats.IncrementStatistics(Statistics::LARGER_TABLE_WRITE_HITS);
				++larger_table_hits;
			}
#endif /* ADAPTIVE_LOCKING */

			return log_entry;
		}

		// someone locked it in the meantime
		// return last element back to the log
		write_log.delete_last();
		
		// read version again
		lock_value = (WriteLock)atomic_load_acquire(write_lock);
		locked = is_write_locked(lock_value);
		YieldCPU();
	}

	// this can never happen
	return NULL;
}

inline void wlpdstm::TxMixinv::WriteWord(void *address, void *value) {
#ifdef MEASURE_TIMING
	uint64_t start;
	bool should_measure_time = ShouldMeasureWriteTiming();

	if(should_measure_time) {
		start = get_clock_count();
		atomic_barrier_full();
	}
#endif /* MEASURE_TIMING */
	
	// map address to the lock
	VersionLock *write_lock = map_address_to_write_lock(address);
	
	// try to lock the address - it will abort if address cannot be locked
	WriteLogEntry *log_entry = LockMemoryStripe(write_lock, address);
	
	// insert (address, value) pair into the log
	log_entry->InsertWordLogEntry(address, value);

#ifdef MEASURE_TIMING
	if(should_measure_time) {
		atomic_barrier_full();
		uint64_t end = get_clock_count();
		write_timing_total += end - start;
		++write_timing_count;
	}
#endif /* MEASURE_TIMING */	
}

#ifdef SUPPORT_LOCAL_WRITES
inline void wlpdstm::TxMixinv::WriteWordLocal(void *address, void *val) {
	WriteWordLocalLogEntry *log_entry = write_local_log.get_next();
	
	log_entry->address = address;
	log_entry->value = *(void **)address;
	
	// update location
	*(void **)address = val;
}
#endif /* SUPPORT_LOCAL_WRITES*/

inline wlpdstm::TxMixinv::WriteWordLogEntry *wlpdstm::TxMixinv::WriteLogEntry::FindWordLogEntry(void *address) {
#ifdef WORD_LOG_HASHTABLE
	unsigned index = HashAddress((uintptr_t)address);

	if(table[index].address == address) {
		return table + index;
	}

	WriteWordLogEntry *curr = table[index].next;
#else
	WriteWordLogEntry *curr = head;
#endif /* WORD_LOG_HASHTABLE */
	
	while(curr != NULL) {
		if(curr->address == address) {
			break;
		}
		
		curr = curr->next;
	}
	
	return curr;
}

inline void wlpdstm::TxMixinv::WriteLogEntry::InsertWordLogEntry(void *address, void *value) {
#ifdef WORD_LOG_HASHTABLE
	unsigned index = HashAddress((uintptr_t)address);

	if(table[index].address == address) {
		table[index].value = value;
	} else if(table[index].address == NULL) {
		table[index].address = address;
		table[index].value = value;
	} else {
		WriteWordLogEntry *curr = table[index].next;

		while(curr != NULL) {
			if(curr->address == address) {
				curr->value = value;
				return;
			}

			curr = curr->next;
		}

		curr = owner->write_word_log_mem_pool.get_next();
		curr->address = address;
		curr->value = value;
		curr->next = table[index].next;
		table[index].next = curr;
	}
#else
	WriteWordLogEntry *entry = FindWordLogEntry(address);
	
	if(entry == NULL) {
		entry = owner->write_word_log_mem_pool.get_next();
		entry->address = address;
		entry->next = head;
		head = entry;
	}
	
	entry->value = value;
#endif /* WORD_LOG_HASHTABLE */
}

inline void wlpdstm::TxMixinv::WriteLogEntry::ClearWordLogEntries() {
#ifdef WORD_LOG_HASHTABLE
	memset(&table, NULL, sizeof(table));
#else
	head = NULL;
#endif /* WORD_LOG_HASHTABLE */
}

inline void *wlpdstm::TxMixinv::ReadWord(void *address) {
#ifdef MEASURE_TIMING
	uint64_t start;
	bool should_measure_time = ShouldMeasureReadTiming();
	
	if(should_measure_time) {
		start = get_clock_count();
		atomic_barrier_full();
	}
#endif /* MEASURE_TIMING */

	void *ret = ReadWordInner(address);

#ifdef MEASURE_TIMING
	if(should_measure_time) {
		atomic_barrier_full();
		uint64_t end = get_clock_count();
		read_timing_total += end - start;
		++read_timing_count;
	}
#endif /* MEASURE_TIMING */

	return ret;
}

inline void *wlpdstm::TxMixinv::ReadWordInner(void *address) {
#ifdef ADAPTIVE_LOCKING
	++reads;
#endif /* ADAPTIVE_LOCKING */
	WriteLock *write_lock = map_address_to_write_lock(address);
	WriteLogEntry *log_entry = (WriteLogEntry *)atomic_load_no_barrier(write_lock);

	// if locked by me return quickly
	if(LockedByMe(log_entry)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)*write_lock;
		WriteWordLogEntry *word_log_entry = log_entry->FindWordLogEntry(address);

		if(word_log_entry != NULL) {
			// if it was written return from log
			return word_log_entry->value;
		} else {
			// if it was not written return from memory
			return (void *)atomic_load_no_barrier(address);
		}		
	}

#ifdef ADAPTIVE_LOCKING
//	if(larger_lock_extent_table.Contains(map_address_to_index(address, larger_lock_extent))) {
//		stats.IncrementStatistics(Statistics::LARGER_TABLE_READ_HITS);
//		++larger_table_read_hits;
//	}
#endif /* ADAPTIVE_LOCKING */
	

	VersionLock *read_lock = map_write_lock_to_read_lock(write_lock);
	VersionLock version = (VersionLock)atomic_load_acquire(read_lock);
	void *value;

	while(true) {
		if(is_read_locked(version)) {
			version = (VersionLock)atomic_load_acquire(read_lock);
			YieldCPU();
			continue;
		}

		value = (void *)atomic_load_acquire(address);
		VersionLock version_2 = (VersionLock)atomic_load_acquire(read_lock);

		if(version != version_2) {
			version = version_2;
			YieldCPU();
			continue;
		}

		ReadLogEntry *entry = read_log.get_next();
		entry->read_lock = read_lock;
		entry->version = version;		

		if(get_value(version) > valid_ts) {
			if(!Extend()) {
				// need to restart here
				stats.IncrementStatistics(Statistics::ABORT_READ_VALIDATE);
#ifdef ADAPTIVE_LOCKING
				++aborts;
#endif /* ADAPTIVE_LOCKING */
				TxRestart(RESTART_VALIDATION);
			}
		}

		break;
	}
	
	return value;
}

// this is validate that occurs during reads and writes not during commit
inline bool wlpdstm::TxMixinv::Validate() {
	ReadLog::iterator iter;
	
	for(iter = read_log.begin();iter.hasNext();iter.next()) {
		ReadLogEntry &entry = *iter;
		VersionLock currentVersion = (VersionLock)atomic_load_no_barrier(entry.read_lock);
		
		if(currentVersion != entry.version) {
			return false;
		}
	}
	
	return true;
}

// validate invoked at commit time
inline bool wlpdstm::TxMixinv::ValidateCommit() {
	ReadLog::iterator iter;

	for(iter = read_log.begin();iter.hasNext();iter.next()) {
		ReadLogEntry &entry = *iter;
		VersionLock currentVersion = (VersionLock)atomic_load_no_barrier(entry.read_lock);
		
		if(currentVersion != entry.version) {
			if(is_read_locked(currentVersion)) {
				WriteLock *write_lock = map_read_lock_to_write_lock(entry.read_lock);
				WriteLogEntry *log_entry = (WriteLogEntry *)atomic_load_no_barrier(write_lock);
				
				if(LockedByMe(log_entry)) {
					continue;
				}
			}
			
			return false;
		}
	}
	
	return true;
}

inline bool wlpdstm::TxMixinv::Extend() {
	unsigned ts = commit_ts.readCurrentTsAcquire();

	if(Validate()) {
		valid_ts = ts;
#ifdef DETAILED_STATS
		stats.IncrementStatistics(Statistics::EXTEND_SUCCESS);
#endif /* DETAILED_STATS */
		return true;
	}

#ifdef DETAILED_STATS
	stats.IncrementStatistics(Statistics::EXTEND_FAILURE);
#endif /* DETAILED_STATS */

	return false;
}

// this function knows maping from addresses to locks
inline void wlpdstm::TxMixinv::LockMemoryBlock(void *address, size_t size) {
//	// lock the first lock
//	VersionLock *first_lock = map_address_to_write_lock(address);
//	stats.IncrementStatistics(Statistics::MEMORY_DEALLOC_LOCKS);
//	LockMemoryStripe(first_lock, address);
//
//	// find the beginning of the next word
//	uintptr_t curr = (uintptr_t)address;
//	uintptr_t end = (uintptr_t)address + size;
//	VersionLock *lock;
//
//	while((lock = map_address_to_write_lock((void *)++curr)) == first_lock) {
//		// skip all addresses covered the first lock
//	}
//
//	unsigned lock_stripe_size = 1 << LOCK_EXTENT;
//
//	for(;curr < end;curr += lock_stripe_size) {
//		stats.IncrementStatistics(Statistics::MEMORY_DEALLOC_LOCKS);
//		LockMemoryStripe(map_address_to_write_lock((void *)curr), (void *)curr);
//	}

//
// Old version
//
	uintptr_t start = (uintptr_t)address;
	uintptr_t end = start + size;
	VersionLock *old = NULL;
	VersionLock *curr;
	
	for(uintptr_t address = start;address < end;address++) {
		curr = map_address_to_write_lock((void *)address);
		
		if(curr != old) {
			LockMemoryStripe(curr, (void *)address);
			old = curr;
		}
	}
}

inline void *wlpdstm::TxMixinv::TxMalloc(size_t size) {
	return mm.TxMalloc(size);
}

inline void wlpdstm::TxMixinv::TxFree(void *ptr, size_t size) {
#ifdef DETAILED_STATS
	stats.IncrementStatistics(Statistics::MEMORY_DEALLOC_COUNT);
	stats.IncrementStatistics(Statistics::MEMORY_DEALLOC_SIZE, size);
#endif /* DETAILED_STATS */
	LockMemoryBlock(ptr, size);
	mm.TxFree(ptr);
}

inline bool wlpdstm::TxMixinv::LockedByMe(WriteLogEntry *log_entry) {
	// this is much faster than going through the log or checking address ranges
	return log_entry != NULL && log_entry->owner == this;
}

inline void wlpdstm::TxMixinv::TxRestart(RestartCause cause) {
	Rollback();
	atomic_store_release(&executing, false);
#ifdef WAIT_ON_SUCC_ABORTS
	if(cause == RESTART_LOCK) {
		if(++succ_aborts > SUCC_ABORTS_MAX) {
			succ_aborts = SUCC_ABORTS_MAX;
		}
		
		if(++succ_aborts >= SUCC_ABORTS_THRESHOLD) {
			WaitOnAbort();
		}
	}
#endif /* WAIT_ON_SUCC_ABORTS */	
	
	RestartJump();
}

inline void wlpdstm::TxMixinv::RestartJump() {
#ifdef USE_LONG_JUMP
	siglongjmp(start_buf, LONG_JMP_RESTART_FLAG);
#else
	throw RestartedException();
#endif /* USE_LONG_JUMP */	
}

inline void wlpdstm::TxMixinv::TxAbort() {
	Rollback();
	atomic_store_release(&executing, false);
	AbortJump();
}

inline void wlpdstm::TxMixinv::AbortJump() {
#ifdef USE_LONG_JUMP
	siglongjmp(start_buf, LONG_JMP_ABORT_FLAG);
#else
	throw AbortedException();
#endif /* USE_LONG_JUMP */
}


inline bool wlpdstm::TxMixinv::ShouldAbortWrite(WriteLock *write_lock) {
	if(aligned_aborted_externally.aborted_externally) {
		return true;
	}
	
	if(aligned_greedy_ts.greedy_ts == MINIMUM_TS) {
		return true;
	}

	WriteLock lock_value = (WriteLock)atomic_load_no_barrier(write_lock);
	
	if(is_write_locked(lock_value)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)lock_value;
		TxMixinv *owner = log_entry->owner;
		
		if(CMStrongerThan(owner)) {
			if(!owner->aligned_aborted_externally.aborted_externally) {
				owner->aligned_aborted_externally.aborted_externally = true;
#ifdef DETAILED_STATS
				stats.IncrementStatistics(Statistics::CM_DECIDE);
#endif /* DETAILED_STATS */
			}
			
			return false;
		}
		
		return true;
	}
	
	return false;
}

inline bool wlpdstm::TxMixinv::CMStrongerThan(TxMixinv *other) {
	if(aligned_greedy_ts.greedy_ts == MINIMUM_TS) {
		return false;
	}
	
	unsigned other_ts = other->aligned_greedy_ts.greedy_ts;
	return other_ts == MINIMUM_TS || aligned_greedy_ts.greedy_ts < other_ts;
}

inline void wlpdstm::TxMixinv::CmStartTx() {
#ifdef SIMPLE_GREEDY
	if(succ_aborts == 0) {
		cm_phase = CM_PHASE_GREEDY;
		aligned_greedy_ts.greedy_ts = cm_ts.getNextTs() + 1;
	}
#else /* two phase greedy */
	cm_phase = CM_PHASE_INITIAL;
	aligned_greedy_ts.greedy_ts = MINIMUM_TS;
	locations_accessed = 0;
#endif /* SIMPLE_GREEDY */
}

inline void wlpdstm::TxMixinv::CmOnAccess() {
#ifndef SIMPLE_GREEDY
	if(cm_phase == CM_PHASE_INITIAL) {
		if(++locations_accessed > CM_ACCESS_THRESHOLD) {
#ifdef DETAILED_STATS
			stats.IncrementStatistics(Statistics::SWITCH_TO_SECOND_CM_PHASE);
#endif /* DETAILED_STATS */
			cm_phase = CM_PHASE_GREEDY;
			aligned_greedy_ts.greedy_ts = cm_ts.getNextTs() + 1;
		}
	}
#endif /* SIMPLE_GREEDY */
}

#ifdef WAIT_ON_SUCC_ABORTS
inline void wlpdstm::TxMixinv::WaitOnAbort() {
	uint64_t cycles_to_wait = random.Get() % (succ_aborts * WAIT_CYCLES_MULTIPLICATOR);
	wait_cycles(cycles_to_wait);
#ifdef DETAILED_STATS
	stats.IncrementStatistics(Statistics::WAIT_ON_ABORT);
#endif /* DETAILED_STATS */
}
#endif /* WAIT_ON_SUCC_ABORTS */

inline void wlpdstm::TxMixinv::InitLastObservedTs() {
	aligned_last_observed_ts.last_observed_ts = MINIMUM_TS;
	start_count = 0;	
}

inline void wlpdstm::TxMixinv::Register() {
	// add itself to the transaction array
	transactions[tid.Get()] = this;
	
	// update thread count
	unsigned my_count = tid.Get() + 1;
	unsigned curr_count = atomic_load_acquire(&thread_count);
	
	while(my_count > curr_count) {
		if(atomic_cas_full(&thread_count, curr_count, my_count)) {
			break;
		}
		
		curr_count = atomic_load_acquire(&thread_count);
	}
}

inline void wlpdstm::TxMixinv::UpdateLastObservedTs(unsigned ts) {
	if(++start_count >= UPDATE_LOCAL_LAST_OBSERVED_TS_FREQUENCY) {
		aligned_last_observed_ts.last_observed_ts = ts - 1;
		start_count = 0;
	}
}

inline unsigned wlpdstm::TxMixinv::UpdateMinimumObservedTs() {
	unsigned ret = MINIMUM_TS;
	
	if(minimum_observed_ts_lock.try_lock()) {
		if(transactions[0]) {
			unsigned minimum_ts = transactions[0]->aligned_last_observed_ts.last_observed_ts;
			
			for(unsigned i = 1;i < thread_count;i++) {
				if(transactions[i] != NULL) {
					unsigned ts = transactions[i]->aligned_last_observed_ts.last_observed_ts;
					
					if(ts < minimum_ts) {
						minimum_ts = ts;
					}
				}
			}
			
			minimum_observed_ts.val = minimum_ts;
			ret = minimum_ts;
		}
		
		minimum_observed_ts_lock.release();
	} else {
		ret = minimum_observed_ts.val;
	}
	
	return ret;
}

inline unsigned wlpdstm::TxMixinv::GetMinimumObservedTs() {
	return minimum_observed_ts.val;
}

inline void wlpdstm::TxMixinv::YieldCPU() {
#ifdef USE_PREEMPTIVE_WAITING
	pre_yield();
#endif	
}

#include <iostream>

inline void wlpdstm::TxMixinv::PrintStatistics() {
#ifdef COLLECT_STATS
	std::cout << std::endl;
	std::cout << "STM internal statistics: " << std::endl;
	std::cout << "========================" << std::endl;
	
	// collect stats in a single collection
	ThreadStatisticsCollection stat_collection;
	
	for(unsigned i = 0;i < thread_count;i++) {
		// these should all be initialized at this point
		stat_collection.Add(&(transactions[i]->stats));
		std::cout << "Thread " << i + 1 << ": " << std::endl;
		transactions[i]->stats.Print(std::cout, 1);
		std::cout << std::endl;

//		std::cout << "\tFalse sharing history: [";
//
//		for(unsigned j = 0;j < FALSE_SHARING_HISTORY_SIZE;j++) {
//			std::cout << transactions[i]->false_sharing_history[j] << " ";
//		}
//
//		std::cout << "]" << std::endl;
	}
	
	std::cout << "Total stats: " << std::endl;
	ThreadStatistics total_stats = stat_collection.MergeAll();
	total_stats.Print(std::cout, 1);
	std::cout << std::endl << "Configuration:" << std::endl;
//	std::cout << "\tLockExtent: " << lock_extent.val << std::endl;
	std::cout << "\tLockExtent: " << LOCK_EXTENT << std::endl;
#endif /* COLLECT_STATS */

#ifdef MEASURE_TIMING
	for(unsigned i = 0;i < thread_count;i++) {
		transactions[i]->PrintTimingData();
	}
#endif /* MEASURE_TIMING */
}

inline bool wlpdstm::TxMixinv::StartSynchronization() {
	if(!atomic_cas_acquire(&synchronization_in_progress.val, false, true)) {
		Synchronize();
		return false;
	}

	for(unsigned i = 0;i < thread_count;i++) {
		while(atomic_load_acquire(&transactions[i]->executing)) {
			// do nothing
		}
	}

	return true;
}

inline void wlpdstm::TxMixinv::EndSynchronization() {
	atomic_store_release(&synchronization_in_progress.val, false);
}

inline bool wlpdstm::TxMixinv::Synchronize() {
	bool ret = false;

	while(atomic_load_acquire(&synchronization_in_progress.val)) {
		ret = true;
	}

	return true;
}

inline void wlpdstm::TxMixinv::RestartCommitTS() {
	commit_ts.restart();
	InitializeReadLocks();
}

#ifdef ADAPTIVE_LOCKING
inline bool wlpdstm::TxMixinv::IsFalseConflict(WriteLogEntry *log_entry, void *address) {
	WriteWordLogEntry *word_log_entry = log_entry->head;
	
	while(word_log_entry != NULL) {
		if(word_log_entry->address == address) {
			return false;
		}

		word_log_entry = word_log_entry->next;
	}

	return true;
}

inline void wlpdstm::TxMixinv::ClearLocalStatistics() {
	writes = 0;
	reads = 0;
	new_writes = 0;
	larger_table_hits = 0;
	aborts = 0;
	false_sharing = 0;
}

inline void wlpdstm::TxMixinv::ApplyTableConfigurationChange() {
	if(proposed_lock_extent.val == LOCK_EXTENT_CHANGED) {
		if(lock_extent.val + 1 < MAX_LOCK_EXTENT_SIZE) {
			larger_lock_extent = lock_extent.val + 1;
		}

		ClearLocalStatistics();
		change_table_configuration_counter = 0;
		old_lock_extent = lock_extent.val;
		proposed_lock_extent.val = NOT_PROPOSING_LOCK_EXTENT;
	}
}

inline void wlpdstm::TxMixinv::ProposeTableChange() {
	if(++change_table_configuration_counter > CHANGE_TABLE_CONFIGURATION_FREQUENCY) {
		ProposeTableChangeInner();
		change_table_configuration_counter = 0;
	}
}

inline void wlpdstm::TxMixinv::ProposeTableChangeInner() {
	return;
/*
	bool proposing = proposed_lock_extent.val != NOT_PROPOSING_LOCK_EXTENT;
	unsigned my_proposition;

	if(!proposing) {
		bool enough_data_aborts = EnoughResizeTableDataAborts();
		bool enough_data_writes = EnoughResizeTableDataWrites();

		if(!enough_data_aborts && !enough_data_writes) {
			return;
		}

		if(enough_data_aborts && lock_extent.val > MIN_LOCK_EXTENT_SIZE) {
			if(CalculateFalseAbortRate() > LOCK_EXTENT_DECREASE_THRESHOLD) {
				my_proposition = lock_extent.val - 1;
				proposing = true;
			}
		}

		if(enough_data_writes && !proposing && lock_extent.val < MAX_LOCK_EXTENT_SIZE) {
			if(CalculateLargerTableHitRate() > LOCK_EXTENT_INCREASE_THRESHOLD
					&& GetFalseSharingHistory(larger_lock_extent) < LOCK_EXTENT_DECREASE_THRESHOLD) {
				my_proposition = larger_lock_extent;
				proposing = true;
			}
		}
	}

	// if no need to change, but there is enough statistical data
	if(!proposing) {
		// increase larger table lock extent to avoid local minima
		if(larger_lock_extent < MAX_LOCK_EXTENT_SIZE &&
				GetFalseSharingHistory(larger_lock_extent + 1) < LOCK_EXTENT_DECREASE_THRESHOLD) {
			++larger_lock_extent;
			ClearLocalStatistics();
		}

		return;
	}

	// check if enough threads want this change
	unsigned threads_wanting_change = 0;

	for(unsigned i = 0;i < thread_count;i++) {
		if(transactions[i] && transactions[i]->proposed_lock_extent.val == my_proposition) {
			++threads_wanting_change;
		}
	}

	if(threads_wanting_change + 1 < (thread_count + 1) / 2) {
		return;
	}

	// try to start synchronization
	if(!StartSynchronization()) {
		ApplyTableConfigurationChange();
		return;
	}

	// now perform the actual change
	stats.IncrementStatistics(Statistics::LOCK_TABLE_RECONFIGURATIONS);
	lock_extent.val = my_proposition;

	// inform everyone of the change
	for(unsigned i = 0;i < thread_count;i++) {
		if(transactions[i]) {
			transactions[i]->proposed_lock_extent.val = LOCK_EXTENT_CHANGED;
		}
	}	

	EndSynchronization();

	// accept the change myself
	ApplyTableConfigurationChange();
*/
}

inline unsigned wlpdstm::TxMixinv::CalculateFalseAbortRate() {
	return aborts ? false_sharing * 100 / aborts : 0;
}

inline unsigned wlpdstm::TxMixinv::CalculateLargerTableHitRate() {
	return larger_table_hits * 100 / writes;
}

//inline unsigned wlpdstm::TxMixinv::GetFalseSharingHistoryIndex(unsigned le) {
//	return le - MIN_LOCK_EXTENT_SIZE;
//}

//inline bool wlpdstm::TxMixinv::EnoughResizeTableDataAborts() {
//	return aborts > SIGNIFICANT_STAT_COUNT_ABORTS;// && writes > SIGNIFICANT_STAT_COUNT_WRITES;
//	return writes > SIGNIFICANT_STAT_COUNT_WRITES;
//}

inline bool wlpdstm::TxMixinv::EnoughResizeTableDataWrites() {
	return writes > SIGNIFICANT_STAT_COUNT_WRITES;
}

//inline unsigned wlpdstm::TxMixinv::GetFalseSharingHistory(unsigned le) {
//	return false_sharing_history[GetFalseSharingHistoryIndex(le)];
//}
#endif /* ADAPTIVE_LOCKING */

#ifdef MEASURE_TIMING
inline bool wlpdstm::TxMixinv::ShouldMeasureWriteTiming() {
	if(++measure_write_timing_count > MEASURE_TIMING_FREQUENCY) {
		measure_write_timing_count = 0;
		return true;
	}

	return false;
}

inline bool wlpdstm::TxMixinv::ShouldMeasureReadTiming() {
	if(++measure_read_timing_count > MEASURE_TIMING_FREQUENCY) {
		measure_read_timing_count = 0;
		return true;
	}
	
	return false;

//	return (get_clock_count() & 0x7F) == 0x7F;

//	uint32_t esp = get_esp();
//	std::cout << esp << std::endl;
//	return (esp & 0x7C) == 0x7C;
}

inline void wlpdstm::TxMixinv::ResetTimingData() {
	measure_write_timing_count = 0;
	measure_read_timing_count = 0;

	write_timing_count = 0;
	write_timing_total = 0;

	read_timing_count = 0;
	read_timing_total = 0;
}

inline void wlpdstm::TxMixinv::PrintTimingData() {
	std::cout << "Average write duration: " << (float)write_timing_total / write_timing_count << std::endl;
	std::cout << "Average read duration: " << (float)read_timing_total / read_timing_count << std::endl;
}
#endif /* MEASURE_TIMING */
