/**
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#include <cstdlib>
#include <stdint.h>

#include "aborted_exc.h"
#include "common/version_lock.h"
#include "constants.h"
#include "common/log.h"
#include "common/timestamp.h"
#include "memory.h"
#include "tid.h"
#include "common/tls.h"
#include "common/cache_aligned_alloc.h"
#include "common/timing.h"

#ifdef COLLECT_STATS
#include "stats.h"
#endif /* COLLECT_STATS */

#ifdef USE_PREEMPTIVE_WAITING
#include "common/preemptive_utils.h"
#endif // USE_PREEMPTIVE_WAITING

#define ADDRESS_SPACE_SIZE 32
#define LOCK_EXTENT_SIZE 6
#define LOCK_OVERLOADING 0

#define VERSION_LOCK_TABLE_SIZE (1 << (ADDRESS_SPACE_SIZE - LOCK_EXTENT_SIZE - LOCK_OVERLOADING))
#define FULL_VERSION_LOCK_TABLE_SIZE (VERSION_LOCK_TABLE_SIZE << 1)

//#define CM_PHASE_THRESHOLD 1

#define CM_ACCESS_THRESHOLD 20

//#define UPDATE_LOCAL_LAST_OBSERVED_TS_FREQUENCY 10
#define UPDATE_LOCAL_LAST_OBSERVED_TS_FREQUENCY 1

namespace wlpdstm {

	union PaddedUnsigned {
		volatile unsigned val;
		char padding[CACHE_LINE_SIZE_BYTES];
	};

	class TxMixinv : public CacheAlignedAlloc {

		//////////////////////////////////
		// log related structures start //
		//////////////////////////////////

		struct ReadLogEntry {
			VersionLock *address;
			VersionLock version;
		};

		struct WriteWordLogEntry {
			void *address;
			void *value;
			WriteWordLogEntry *next;
		};

		struct WriteWordLocalLogEntry {
			void *address;
			void *value;
		};

		struct WriteLogEntry {
			VersionLock *address;

			TxMixinv *owner;

			WriteWordLogEntry *head;

			// TODO: perhaps add a bitmap to speed up lookups
			// might not be needed as memory regions are short

			// methods
			void InsertWordLogEntry(void *address, void *value);

			WriteWordLogEntry *FindWordLogEntry(void *address);
		};

		typedef Log<ReadLogEntry> ReadLog;
		typedef Log<WriteLogEntry> WriteLog;
		typedef Log<WriteWordLocalLogEntry> WriteLocalLog;
		typedef Log<WriteWordLogEntry> WriteWordLogMemPool;

		////////////////////////////////
		// log related structures end //
		////////////////////////////////

		// transaction state
		public:
			enum TxState {
				TX_STATE_IDLE,
				TX_STATE_RUNNING,
				TX_STATE_COMMITTING
			};

			enum CmPhase {
				CM_PHASE_INITIAL,
				CM_PHASE_GREEDY
			};

		public:
			static void GlobalInit();

			void ThreadInit();

			void TxStart();

			void TxCommit();

			void TxAbort();

			void TxAbortFull();

			void WriteWord(void *address, void *val);

			void WriteWordLocal(void *address, void *val);

			void *ReadWord(void *address);

			void *TxMalloc(size_t size);

			void TxFree(void *ptr, size_t size);

			void TxRestart();

		private:
			void LockMemoryBlock(void *address, size_t size);

			static unsigned map_address_to_index(void *address);

			static VersionLock *map_address_to_lock(void *address);

			static VersionLock *map_address_to_old_version(void *address);

			static VersionLock *map_lock_to_old_version(VersionLock *lock_address);

			WriteLogEntry *LockMemoryStripe(VersionLock *lock_address);

			void Rollback();

			bool Validate();

			bool Extend();

			bool Extend(unsigned ts);

			bool LockedByMe(WriteLogEntry *log_entry);

			///////////////////////////
			// contention management //
			///////////////////////////

			bool ShouldAbortWrite(VersionLock *lock_address);

			// TODO: isolate CM in a separate class
			bool CMStrongerThan(TxMixinv *other);

//			bool Retrying() const;

//			bool SecondCMPhase() const;

			void CmStartTx();

			void CmOnAccess();

			void InitLastObservedTs();

			void UpdateLastObservedTs(unsigned ts);

			void Register();

		public:
			static unsigned UpdateMinimumObservedTs();

			static unsigned GetMinimumObservedTs();

		private:
			// local
			ReadLog read_log;

			// local
			WriteLog write_log;

			// local
			WriteLocalLog write_local_log;

			// local
			WriteWordLogMemPool write_word_log_mem_pool;

			// local
			unsigned valid_ts;

			// local
//			unsigned succ_abort_cnt;

			// contention management
			// local
			unsigned locations_accessed;

			// local
			CmPhase cm_phase;

			// local
			unsigned start_count;

			//////////////////////
			// shared variables //
			//////////////////////

			static VersionLock version_lock_table[FULL_VERSION_LOCK_TABLE_SIZE];

//			static VersionLock old_version_table[VERSION_LOCK_TABLE_SIZE];

			static GlobalTimestamp commit_ts;

			static GlobalTimestamp cm_ts;

			static PaddedUnsigned minimum_observed_ts;

			static PaddedSpinTryLock minimum_observed_ts_lock;

			static TxMixinv *transactions[MAX_THREADS];

			static unsigned thread_count;

			///////////////
			// constants //
			///////////////

			//static const VersionLock MINIMUM_VERSION;

		public:
			// local
			Tid tid;

			// local
			MemoryManager mm;

		private:
			// data accessed by other tx
			// ro shared
			CACHE_LINE_ALIGNED union {
				volatile TxState state;
				char padding[CACHE_LINE_SIZE_BYTES];
			} aligned_state;

			// ro shared
			CACHE_LINE_ALIGNED union {
				volatile unsigned greedy_ts;
				char padding[CACHE_LINE_SIZE_BYTES];
			} aligned_greedy_ts;

			// w shared
			CACHE_LINE_ALIGNED union {
				volatile bool aborted_externally;
				char padding[CACHE_LINE_SIZE_BYTES];
			} aligned_aborted_externally;

			// r shared
			CACHE_LINE_ALIGNED union {
				volatile unsigned last_observed_ts;
				char padding[CACHE_LINE_SIZE_BYTES];
			} aligned_last_observed_ts;

#ifdef COLLECT_STATS
		// stats
		private:
			// local
			ThreadStatistics stats;
#endif /* COLLECT_STATS */

		public:
			static void PrintStatistics();

	} __attribute__ ((aligned(CACHE_LINE_SIZE_BYTES)));
}

inline void wlpdstm::TxMixinv::GlobalInit() {
	// initialize object locks
	VersionLock initial_version = get_version_lock(MINIMUM_TS + 1);

	// initialize global table lock
	for(int i = 0;i < FULL_VERSION_LOCK_TABLE_SIZE;i++) {
		version_lock_table[i] = initial_version;
	}

	// initialize shared data
	minimum_observed_ts.val = MINIMUM_TS;
	
	// initialize memory manager
	MemoryManager::GlobalInit();
}

// TODO: the most natural way would be to map addresses to lock based
// on objects that addresses point to, but this is not so simple
inline unsigned wlpdstm::TxMixinv::map_address_to_index(void *address) {
	return (((uintptr_t)address >> (LOCK_EXTENT_SIZE)) & (VERSION_LOCK_TABLE_SIZE - 1)) << 1;
}

inline wlpdstm::VersionLock *wlpdstm::TxMixinv::map_address_to_lock(void *address) {
	return version_lock_table + map_address_to_index(address);
}

inline wlpdstm::VersionLock *wlpdstm::TxMixinv::map_address_to_old_version(void *address) {
	return map_address_to_lock(address) + 1;
}

inline wlpdstm::VersionLock *wlpdstm::TxMixinv::map_lock_to_old_version(
		VersionLock *lock_address) {
	return lock_address + 1;
}

inline void wlpdstm::TxMixinv::ThreadInit() {
	aligned_aborted_externally.aborted_externally = false;
	aligned_state.state = TX_STATE_IDLE;

	InitLastObservedTs();

	Register();

	// locally initialize memory manager
	mm.ThreadInit(tid.Get());
}

inline void wlpdstm::TxMixinv::TxStart() {
	mm.TxStart();

	// get validity timestamp
	valid_ts = commit_ts.readCurrentTs();

	UpdateLastObservedTs(valid_ts);

	CmStartTx();

	// reset aborted flag
	aligned_aborted_externally.aborted_externally = false;

	// start running
	atomic_store_full(&aligned_state.state, TX_STATE_RUNNING);
}

inline void wlpdstm::TxMixinv::TxCommit() {
	unsigned ts = MINIMUM_TS;

	if(!write_log.empty()) {
		// first change state
		aligned_state.state = TX_STATE_COMMITTING;
//		atomic_store_full(&aligned_state.state, TX_STATE_COMMITTING);

		// increase global timestamp
		ts = commit_ts.getNextTsFull() + 1;

		if(ts != valid_ts + 1 && !Validate()) {
			TxRestart();
		}

		// generate new version in correct format
		VersionLock commitVersion = get_version_lock(ts);

		// process redo log
		for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
			WriteLogEntry &entry = *iter;
			*(map_lock_to_old_version(entry.address)) = MINIMUM_VERSION;
		}

		atomic_barrier_full();

		for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
			WriteLogEntry &entry = *iter;

			// now update actual values
			WriteWordLogEntry *word_log_entry = entry.head;

			while(word_log_entry != NULL) {
				*(void **)word_log_entry->address =
					word_log_entry->value;
				word_log_entry = word_log_entry->next;
			}

			// release the lock
			atomic_store_release(entry.address, commitVersion);
			*map_lock_to_old_version(entry.address) = commitVersion;
		}
	}

	// finish committing values
//	aligned_state.state = TX_STATE_IDLE;
	atomic_store_release(&aligned_state.state, TX_STATE_IDLE);

	write_log.clear();
	read_log.clear();
	write_local_log.clear();
	write_word_log_mem_pool.clear();

	// commit mem
	mm.TxCommit<TxMixinv>(ts);

#ifdef COLLECT_STATS
	stats.stats->Increment(Statistics::COMMIT);
#endif
}

inline void wlpdstm::TxMixinv::Rollback() {
	// rollback local writes
	for(WriteLocalLog::iterator iter = write_local_log.begin();iter.hasNext();iter.next()) {
		WriteWordLocalLogEntry &entry = *iter;
		*(void **)entry.address = entry.value;
	}

	// drop locks
	for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
		WriteLogEntry &entry = *iter;
		*entry.address = *map_lock_to_old_version(entry.address);
	}

	// empty logs
	read_log.clear();
	write_log.clear();
	write_local_log.clear();
	write_word_log_mem_pool.clear();

#ifdef USE_PREEMPTIVE_WAITING
	pre_yield();
#endif

	mm.TxAbort();
}

inline wlpdstm::TxMixinv::WriteLogEntry *wlpdstm::TxMixinv::LockMemoryStripe(VersionLock *lock_address) {
	// read lock value
	VersionLock version = (VersionLock)atomic_load_no_barrier(lock_address);

	if(is_locked(version)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);

		if(LockedByMe(log_entry)) {
			return log_entry;
		}
	}

	while(true) {
		if(is_locked(version)) {
			if(ShouldAbortWrite(lock_address)) {
				TxRestart();
			} else {
				version = (VersionLock)atomic_load_acquire(lock_address);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		// the word read was unlocked, check validity
		// TODO: Validity need not be checked if the word was not read
		// before, but this requires a pass over read set or use of
		// some lookup mechanism (Bloom filter being used in other
		// implementations), for now skip this check and assume all
		// written words were read before. ??? Is this check needed at all?
		//
		// I will implement something later and check if it helps or not.
		//
		// ~1% red-black tree
		//
///////////////////////////////////////////////////////////////////////////////
// BUG: It is possible that this validity check should go below lock acquire //
///////////////////////////////////////////////////////////////////////////////
//		if(get_value(version) > valid_ts) {
//			// TODO: check if extending really helps or just slows down unnecessarily
//			if(!Extend()) {
//				// must abort here
//				TxRestart();
//			}
//		}
/////////////
// BUG END //
/////////////

		// prepare write log entry
		WriteLogEntry *log_entry = write_log.get_next();
		log_entry->address = lock_address;
//		log_entry->old_version = version;
		log_entry->head = NULL; // need this here
		log_entry->owner = this;
//		++log_entry->entry_version;

		// get lock value
		VersionLock locked_entry = get_locked(log_entry);

		// try to cas log_entry into handle
		if(atomic_cas_release(lock_address, version, locked_entry)) {
			if(get_value(version) > valid_ts) {
				// TODO: check if extending really helps or just slows down unnecessarily
				if(!Extend()) {
					// must abort here
					TxRestart();
				}
			}
			
			// success
			CmOnAccess();
			return log_entry;
		} else {
			// return last element back to the log
			write_log.delete_last();

			// read version again
			version = (VersionLock)atomic_load_acquire(lock_address);
#ifdef USE_PREEMPTIVE_WAITING
			pre_yield();
#endif
		}
	}

	// this can never happen
	return NULL;
}

inline void wlpdstm::TxMixinv::WriteWord(void *address, void *value) {
	// map address to the lock
	VersionLock *lock_address = map_address_to_lock(address);

	// try to lock the address - it will abort if address cannot be locked
	WriteLogEntry *log_entry = LockMemoryStripe(lock_address);

	// insert (address, value) pair into the log
	log_entry->InsertWordLogEntry(address, value);
}

inline void wlpdstm::TxMixinv::WriteWordLocal(void *address, void *val) {
	WriteWordLocalLogEntry *log_entry = write_local_log.get_next();

	log_entry->address = address;
	log_entry->value = *(void **)address;

	// update location
	*(void **)address = val;
}

inline wlpdstm::TxMixinv::WriteWordLogEntry *wlpdstm::TxMixinv::WriteLogEntry::FindWordLogEntry(
		void *address) {
	WriteWordLogEntry *curr = head;

	while(curr != NULL) {
		if(curr->address == address) {
			break;
		}

		curr = curr->next;
	}

	return curr;
}

inline void wlpdstm::TxMixinv::WriteLogEntry::InsertWordLogEntry(
		void *address, void *value) {
	WriteWordLogEntry *entry = FindWordLogEntry(address);

	if(entry == NULL) {
		entry = owner->write_word_log_mem_pool.get_next();
		entry->address = address;
		entry->next = head;
		head = entry;
	}

	entry->value = value;
}

inline void *wlpdstm::TxMixinv::ReadWord(void *address) {
	// map address to the lock
	VersionLock *lock_address = map_address_to_lock(address);

	// read lock value
	VersionLock version = (VersionLock)atomic_load_no_barrier(lock_address);

	// if locked by me, return value
	if(is_locked(version)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);

		if(LockedByMe(log_entry)) {
			WriteWordLogEntry *word_log_entry = log_entry->FindWordLogEntry(address);

			if(word_log_entry != NULL) {
				// if it was written return from log
				return word_log_entry->value;
			} else {
				// if it was not written return from memory
				return (void *)atomic_load_no_barrier(address);
			}

			// no need to add this address to read set, because it cannot
			// be changed as it is locked by this tx
		}
	}

//	VersionLock ver;
	void *value;
	bool locked;
	VersionLock old_version;
	WriteLogEntry *log_entry;
	TxMixinv *owner;

	while(true) {
		locked = is_locked(version);

		if(locked) {
			log_entry = (WriteLogEntry *)get_ptr(version);
			owner = log_entry->owner; // this is constant
			TxState owner_state = owner->aligned_state.state;
//			TxState owner_state = (TxSTate)atomic_load_acquire(&owner->aligned_state.state);

			// avoid taking timestamp that is too high
			if(owner_state == TX_STATE_COMMITTING) {
				version = (VersionLock)atomic_load_acquire(lock_address);
				// TODO: wait while it finishes here perhaps
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}

			old_version = (VersionLock)atomic_load_acquire(map_lock_to_old_version(lock_address));

			// ran into middle of object commit
			if(old_version == MINIMUM_VERSION) {
				version = (VersionLock)atomic_load_acquire(lock_address);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		// read value
		value = (void *)atomic_load_acquire(address);
		VersionLock version_2 = (VersionLock)atomic_load_acquire(lock_address);

		if(locked) {
			VersionLock old_version_2 = (VersionLock)atomic_load_acquire(
					map_lock_to_old_version(lock_address));

			if(old_version != old_version_2 || version != version_2) {
				version = version_2;
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}

		} else {
			if(version != version_2) {
				version = version_2;
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		break;
	}

	VersionLock ver = locked ? old_version : version;

	// check validity
	if(get_value(ver) > valid_ts) {
/*		if(locked) {
			unsigned ts = commit_ts.readCurrentTsAcquire();
			TxState owner_state = (TxState)atomic_load_acquire(&owner->aligned_state.state);

			if(owner_state == TX_STATE_COMMITTING || !Extend(ts)) {
				TxRestart();
			}
		} else {*/
			if(!Extend()) {
				// must abort here
				TxRestart();
			}
//		}
	}

	// here all is consistent and valid :)
	ReadLogEntry *entry = read_log.get_next();
	entry->address = lock_address;
	entry->version = ver;
	// TODO: think about this, maybe count only writes
	//CmOnAccess();
	return value;
}

inline bool wlpdstm::TxMixinv::Validate() {
	ReadLog::iterator iter;

	for(iter = read_log.begin();iter.hasNext();iter.next()) {
		ReadLogEntry &entry = *iter;

		VersionLock currentVersion = (VersionLock)atomic_load_no_barrier(
			entry.address);

		if(!is_locked(currentVersion)) {
			if(currentVersion != entry.version) {
				return false;
			}
		} else {
			WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(currentVersion);

			if(LockedByMe(log_entry)) {
				continue;
			}

			TxMixinv *owner = log_entry->owner;
			TxState owner_state = owner->aligned_state.state;

			if(owner_state == TX_STATE_COMMITTING) {
				return false;
			}

			VersionLock old_version = (VersionLock)atomic_load_acquire(
				map_lock_to_old_version(entry.address));

			if(old_version != entry.version) {
				return false;
			}
		}
	}

	return true;
}

inline bool wlpdstm::TxMixinv::Extend() {
	unsigned ts = commit_ts.readCurrentTsAcquire();
	return Extend(ts);
}

inline bool wlpdstm::TxMixinv::Extend(unsigned ts) {
	if(Validate()) {
		valid_ts = ts;
#ifdef COLLECT_STATS
		stats.stats->Increment(Statistics::EXTEND_SUCCESS);
#endif
		return true;
	}

#ifdef COLLECT_STATS
	stats.stats->Increment(Statistics::EXTEND_FAILURE);
#endif

	return false;
}

inline void wlpdstm::TxMixinv::LockMemoryBlock(void *address, size_t size) {
	uintptr_t start = (uintptr_t)address;
	uintptr_t end = start + size;
	VersionLock *old = NULL;
	VersionLock *curr;

	for(uintptr_t address = start;address < end;address++) {
		curr = map_address_to_lock((void *)address);

		if(curr != old) {
			LockMemoryStripe(curr);
			old = curr;
		}
	}
/*
	// calc last address of the block
	void *end_address = (char *)address + size - 1;

	// get start and end version lock address
	// knowledge about how addresses are mapped to locks is used here
	VersionLock *start = map_address_to_lock(address);
	VersionLock *end = map_address_to_lock(end_address);

	// iterate through locks and lock them all
	for(VersionLock *iter = start;iter <= end;iter++) {
		LockMemoryStripe(iter);
	}
*/
}

inline void *wlpdstm::TxMixinv::TxMalloc(size_t size) {
	return mm.TxMalloc(size);
}

inline void wlpdstm::TxMixinv::TxFree(void *ptr, size_t size) {
	LockMemoryBlock(ptr, size);
	mm.TxFree(ptr);
}

inline bool wlpdstm::TxMixinv::LockedByMe(WriteLogEntry *log_entry) {
	// this is faster (at least with two threads) than searching though write log
	// even for red-black tree micro benchmark
	return log_entry->owner == this;
}

inline void wlpdstm::TxMixinv::TxRestart() {
	Rollback();

	aligned_state.state = TX_STATE_IDLE;
//	succ_abort_cnt++;

	throw RestartedException();
}

inline void wlpdstm::TxMixinv::TxAbort() {
	Rollback();
	
	aligned_state.state = TX_STATE_IDLE;
//	succ_abort_cnt = 0;

	throw AbortedException();
}


inline bool wlpdstm::TxMixinv::ShouldAbortWrite(VersionLock *lock_address) {
	if(aligned_aborted_externally.aborted_externally) {
		return true;
	}

	if(aligned_greedy_ts.greedy_ts == MINIMUM_TS) {
		return true;
	}

	VersionLock version = (VersionLock)atomic_load_no_barrier(lock_address);

	if(is_locked(version)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);
		TxMixinv *owner = log_entry->owner;

		if(CMStrongerThan(owner)) {
			if(!owner->aligned_aborted_externally.aborted_externally) {
				owner->aligned_aborted_externally.aborted_externally = true;
			}

			return false;
		}

		return true;
	}

	return false;
}

inline bool wlpdstm::TxMixinv::CMStrongerThan(TxMixinv *other) {
	if(aligned_greedy_ts.greedy_ts == MINIMUM_TS) {
		return false;
	}

	unsigned other_ts = other->aligned_greedy_ts.greedy_ts;
	return other_ts == MINIMUM_TS ||
		(aligned_greedy_ts.greedy_ts < other_ts);
}

inline void wlpdstm::TxMixinv::CmStartTx() {
	cm_phase = CM_PHASE_INITIAL;
	locations_accessed = 0;

	aligned_greedy_ts.greedy_ts = MINIMUM_TS;
}

inline void wlpdstm::TxMixinv::CmOnAccess() {
	if(cm_phase == CM_PHASE_INITIAL) {
		if(++locations_accessed > CM_ACCESS_THRESHOLD) {
			cm_phase = CM_PHASE_GREEDY;
			aligned_greedy_ts.greedy_ts = cm_ts.getNextTs() + 1;
		}
	}
}


//inline bool wlpdstm::TxMixinv::Retrying() const {
//	return succ_abort_cnt != 0;
//}

//inline bool wlpdstm::TxMixinv::SecondCMPhase() const {
//	return succ_abort_cnt == CM_PHASE_THRESHOLD;
//}

inline void wlpdstm::TxMixinv::InitLastObservedTs() {
	aligned_last_observed_ts.last_observed_ts = MINIMUM_TS;
	start_count = 0;	
}

inline void wlpdstm::TxMixinv::UpdateLastObservedTs(unsigned ts) {
	if(++start_count >= UPDATE_LOCAL_LAST_OBSERVED_TS_FREQUENCY) {
		aligned_last_observed_ts.last_observed_ts = ts - 1;
		start_count = 0;
	}
}

inline void wlpdstm::TxMixinv::Register() {
	// add itself to the transaction array
	transactions[tid.Get()] = this;

	// update thread count
	unsigned my_count = tid.Get() + 1;
	unsigned curr_count = atomic_load_acquire(&thread_count);

	while(my_count > curr_count) {
		if(atomic_cas_full(&thread_count, curr_count, my_count)) {
			break;
		}

		curr_count = atomic_load_acquire(&thread_count);
	}
}

inline unsigned wlpdstm::TxMixinv::UpdateMinimumObservedTs() {
	unsigned ret = MINIMUM_TS;

	if(minimum_observed_ts_lock.try_lock()) {
		if(transactions[0]) {
			unsigned minimum_ts =
				transactions[0]->aligned_last_observed_ts.last_observed_ts;

			for(unsigned i = 1;i < thread_count;i++) {
				if(transactions[i] != NULL) {
					unsigned ts =
						transactions[i]->aligned_last_observed_ts.last_observed_ts;

					if(ts < minimum_ts) {
						minimum_ts = ts;
					}
				}
			}

			minimum_observed_ts.val = minimum_ts;
			ret = minimum_ts;
		}

		minimum_observed_ts_lock.release();
	} else {
		ret = minimum_observed_ts.val;
	}

	return ret;
}

inline unsigned wlpdstm::TxMixinv::GetMinimumObservedTs() {
	return minimum_observed_ts.val;
}

#include <iostream>

inline void wlpdstm::TxMixinv::PrintStatistics() {
#ifdef COLLECT_STATS
	std::cout << std::endl;
	std::cout << "STM internal statistics: " << std::endl;
	std::cout << "========================" << std::endl;

	// collect stats in a single collection
	ThreadStatisticsCollection stat_collection;

	for(unsigned i = 0;i < thread_count;i++) {
		// these should all be initialized at this point
		stat_collection.Add(&(transactions[i]->stats));
		std::cout << "Thread " << i + 1 << ": " << std::endl;
		transactions[i]->stats.Print(std::cout, 1);
		std::cout << std::endl;
	}

	std::cout << "Total stats: " << std::endl;
	ThreadStatistics total_stats = stat_collection.MergeAll();
	total_stats.Print(std::cout, 1);
#endif /* COLLECT_STATS */
}
